#include "comp/entity.h"
#include "comp/physical.h"
#include "comp/visible.h"
#include "comp/mechanical.h"
#include "comp/fighter.h"
#include "enemies.h"
#include <string.h>

/* Declare animations for an enemy */
#define ANIMS(name, I, W, A, H, D) \
    .anim_idle =    &anims_ ## name ## _ ## I, \
    .anim_walking = &anims_ ## name ## _ ## W, \
    .anim_attack =  &anims_ ## name ## _ ## A, \
    .anim_hit =     &anims_ ## name ## _ ## H, \
    .anim_death =   &anims_ ## name ## _ ## D

static enemy_t const slime = {
    .name = "Slime",
    ANIMS(slime, Idle, Walking, Idle, Hit, Death),
    .hitbox = (rect){ -fix(3)/16, fix(4)/16, -fix(2)/16, fix(3)/16 },
    .sprite = (rect){ -fix(5)/16, fix(5)/16, -fix(4)/16, fix(3)/16 },
    .limits = {
        .max_speed = fix(1),
        .friction = fix(0.6),
    },
    .HP = {
        .base = fix(10),
        .growth = fix(5),
        .affinity = fix(3),
    },
    .ATK = {
        .base = fix(5),
        .growth = fix(2),
        .affinity = fix(1),
    },
    .DEF = {
        .base = fix(2),
        .growth = fix(1),
        .affinity = fix(0.5),
    },
    .shadow_size = 4,
    .z = 0,
};

static enemy_t const bat = {
    .name = "Bat",
    ANIMS(bat, Idle, Idle, Idle, Hit, Death),
    .hitbox = (rect){ -fix(3)/16, fix(4)/16, -fix(2)/16, fix(3)/16 },
    .sprite = (rect){ -fix(5)/16, fix(5)/16, -fix(4)/16, fix(3)/16 },
    .limits = {
        .max_speed = fix(1.8),
        .friction = fix(0.8),
    },
    .HP = {
        .base = fix(8),
        .growth = fix(4),
        .affinity = fix(2),
    },
    .ATK = {
        .base = fix(8),
        .growth = fix(2),
        .affinity = fix(2),
    },
    .DEF = {
        .base = fix(2),
        .growth = fix(2),
        .affinity = fix(1),
    },
    .shadow_size = 4,
    .z = fix(0.75),
};

static enemy_t const gunslinger = {
    .name = "Gunslinger",
    ANIMS(gunslinger, Idle, Walking, Fire, Hit, Death),
    .hitbox = (rect){ -fix(3)/16, fix(4)/16, -fix(2)/16, fix(3)/16 },
    .sprite = (rect){ -fix(5)/16, fix(5)/16, -fix(4)/16, fix(3)/16 },
    .limits = {
        .max_speed = fix(1.8),
        .friction = fix(0.8),
    },
    .HP = {
        .base = fix(8),
        .growth = fix(4),
        .affinity = fix(2),
    },
    .ATK = {
        .base = fix(8),
        .growth = fix(2),
        .affinity = fix(2),
    },
    .DEF = {
        .base = fix(2),
        .growth = fix(2),
        .affinity = fix(1),
    },
    .shadow_size = 4,
    .z = fix(0.375),
};

enemy_t const * const enemies[] = {
    [ENEMY_SLIME]       = &slime,
    [ENEMY_BAT]         = &bat,
    [ENEMY_GUNSLINGER]  = &gunslinger,
};

static int instantiate_stat(enemy_stat_t const *stat, int level)
{
    fixed_t value = stat->base + stat->growth * (level - 1) +
        stat->affinity * (level * (level - 1) / 2);

    fixed_t variation = value / 8;
    value = (value - variation) + (rand() % (2 * variation));

    return ffloor(value);
}

entity_t *enemy_make(int enemy_id, int level)
{
    if(enemy_id < 0 || (size_t)enemy_id >= sizeof enemies / sizeof *enemies)
        return NULL;

    entity_t *e = entity_make(physical, visible, mechanical, fighter);
    if(e == NULL)
        return NULL;

    enemy_t const *data = enemies[enemy_id];

    /* These will probably be overridden by the caller */

    physical_t *p = getcomp(e, physical);
    p->x = fix(0);
    p->y = fix(0);
    p->hitbox = data->hitbox;
    p->facing = LEFT;

    visible_t *v = getcomp(e, visible);
    v->z = data->z;
    v->sprite_plane = VERTICAL;
    v->shadow_size = data->shadow_size;
    v->anim_priority = 0;
    visible_set_anim(e, data->anim_idle, 1);

    mechanical_t *m = getcomp(e, mechanical);
    m->limits = &data->limits;
    m->vx = fix(0);
    m->vy = fix(0);
    m->dash = fix(0);
    m->dash_facing = LEFT;

    /* Instantiate fighter statistics */
    fighter_t *f = getcomp(e, fighter);
    memset(f, 0, sizeof *f);

    f->identity = enemy_id;
    f->HP_max = instantiate_stat(&data->HP, level);
    f->ATK    = instantiate_stat(&data->ATK, level);
    f->DEF    = instantiate_stat(&data->DEF, level);
    f->HP = f->HP_max;
    f->combo_length = 1;

    return e;
}
